/// Josh's User Interface Library
/// A C++20 Library for creating, styling, and rendering of a UI/UX widgets.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file Widget.hpp
/// @desc Base Widget Class - All JUI widgets extend their behavior from this class.
/// @edit 2024-10-11

#pragma once

#include <JUI/Widgets/TextButton.hpp>

#include "ListLayout.hpp"
#include "Separator.hpp"
#include "JUI/Mixins/Pinnable.hpp"
#include "JUI/Mixins/TextStyleInterface.hpp"

namespace JUI {

    class ContextMenu;

    // TODO: Move to NestableMenu.hpp after resolving circular dependency.

    /// An abstract class that defines the interface for nestable menus - menu objects that contain similar sub-menus.
    class NestableMenu {
        virtual ContextMenu* AddSubmenu(const std::string& text) = 0;
        virtual TextButton* AddButton(const std::string& text) = 0;
        virtual TextButton* AddButton(const std::string& text, const std::function<void()>& callback) = 0;
        virtual Separator* AddSeparator(const UDim& size = 5_px) = 0;
    };

    class RectStyleInterface {
    public:
        virtual Color4 BGColor() const = 0;
        virtual Color4 BorderColor() const = 0;
        virtual enum BorderMode BorderMode() const = 0;
        virtual float BorderWidth() const = 0;
        virtual float CornerRounding() const = 0;

        virtual void BGColor(const Color4& color) = 0;
        virtual void BorderColor(const Color4& color) = 0;
        virtual void BorderMode(const enum BorderMode& borderMode) = 0;

    };



    /// A vertically-descending list menu of button objects, which supports sub-menus much like class UtilityBar.
    /// @see UtilityBar
    class ContextMenu  : public Rect, public Pinnable, public NestableMenu {
    public:

        Event<> PresumeMouseFocusLost;

        // TODO: Work in progress.
        void PropogateTextSize(int value) {
            for (auto* child : children) {
                TextStyleInterface* textual = dynamic_cast<TextStyleInterface *>(child);

                if (textual)
                    textual->TextSize(value);
            }
        }

        /// Sets the TextSize property for the child elements of this widget.
        void TextSize(int value) {
            // TODO: This works fine for the ContextMenu, where children are assuemed to only by submenus and buttons.
            // Some widgets will need to be handled specially, because only some of their children are logically bound to it.
            PropogateTextSize(value);
        }

        ContextMenu();

        explicit ContextMenu(Widget* parent);


        void SetFont(const JGL::Font& use_my_font);

        TextButton* AddButton(const std::string &name) override;
        TextButton* AddButton(const std::string& name, const std::function<void()> &callback) override;
        Separator* AddSeparator(const UDim& size = 5_px) override;
        ContextMenu* AddSubmenu(const std::string& name) override;


        /// Sets whether the widget will automatically hide itself when the mouse leaves its bounding box.
        void CloseOnHoverLoss(bool value);

        /// @return Whether the widget will automatically hide itself when the mouse leaves its bounding box.
        [[nodiscard]] bool CloseOnHoverLoss() const;


        void Update(float delta) override {
            Rect::Update(delta);

            if (parent->IsMouseInside() || IsMouseInside() || IsMouseInsideDescendants())
                mouse_outside_tests = 0;
            else
                mouse_outside_tests++;

            if (mouse_outside_tests >= 3)
                PresumeMouseFocusLost();
        }

    protected:
        VerticalListLayout* layout;
        JGL::Font font;
        bool close_on_hover_loss = true;
        float aggregate_height = 0;
        float estimated_width = 0;
        float mouse_outside_tests = 0;

    private:

    };
}
