#include <JUI/Widgets/TextInputForm.hpp>
#include <ReWindow/InputService.h>

namespace JUI {
    TextInputForm::TextInputForm() : TextRect(), Clickable() {
        Name("TextInputForm");
    }
    TextInputForm::TextInputForm(Widget* parent) : TextInputForm() {
        this->Parent(parent);
    }

    bool TextInputForm::ObserveMouseMovement(const Vector2 &latest_known_pos) {
        return Widget::ObserveMouseMovement(latest_known_pos);
    }

    std::string TextInputForm::GetAutocompleteText() const { return autocomplete_text;}

    void TextInputForm::SetAutoCompleteText(const std::string &text) { autocomplete_text = text;}

    Color4 TextInputForm::GetAutocompleteTextColor() const { return autocomplete_color; }

    void TextInputForm::SetAutocompleteTextColor(const Color4 &color) { autocomplete_color = color; }

    bool TextInputForm::HideAutocompleteOnSelect() const { return hide_autocomplete_on_select; }

    void TextInputForm::SetHideAutocompleteOnSelect(bool hide) { hide_autocomplete_on_select = hide; }

    bool TextInputForm::AutocompleteTextEnabled() const { return autocomplete_text_enabled; }

    void TextInputForm::SetAutocompleteTextEnabled(bool enabled) { autocomplete_text_enabled = enabled; }

    bool TextInputForm::ObserveMouseInput(MouseButton btn, bool pressed) {

        if (pressed && btn == MouseButton::Left) {

            if (IsMouseInside()) {
                if (!focused) {
                    OnSelect.Invoke();
                }
                focused = true;
                return true;
            } else {
                if (focused)
                    OnDeselect.Invoke();
                focused = false;
            }

        }
        return Widget::ObserveMouseInput(btn, pressed);
    }

    void TextInputForm::InnerDraw()
    {

        Rect::InnerDraw();

        Vector2 abs_pos = this->GetAbsolutePosition();
        Vector2 abs_size = this->GetAbsoluteSize();
        // TODO: Apply this everywhere..
        Vector2 pad_shifts_pos_by = {
            pad_left.Pixels + (pad_left.Scale * abs_size.x),
            pad_top.Pixels + (pad_left.Scale * abs_size.y)
        };

        Vector2 pad_shrinks_size_by = {
            pad_right.Pixels + (pad_left.Scale * abs_size.x),
            pad_bottom.Pixels + (pad_right.Scale * abs_size.y)
        };

        Vector2 pos = abs_pos + pad_shifts_pos_by;
        Vector2 size = abs_pos - (pad_shrinks_size_by*2.f);

        if (!focused || !hide_autocomplete_on_select)
            TextBase::Draw(pos, size, autocomplete_text, this->text_size, autocomplete_color);

        TextBase::Draw(pos, size, content, text_size, text_color);

        //TextRect::InnerDraw();
    }

    void TextInputForm::Draw() {
        TextRect::Draw();
    }

    void TextInputForm::Update(float elapsed) {
        TextRect::Update(elapsed);

        // TODO: Make cursor actually blink
        if (focused) {
            cursor_blink_time += elapsed;

            if (cursor_blink_time > 1.f)
                cursor_blink_time = 0.f;

            if (cursor_blink_time > 0.5f) {
                std::string result = input_buffer;

                result.insert(cursor_position, 1, '|');
                SetContent(result);
            } else {
                SetContent(input_buffer);
            }
        }
    }

    std::string uppercase(const std::string& input) {
        std::string s = input;
        std::ranges::transform(s, s.begin(), ::toupper);
        return s;
    }

    std::string lowercase(const std::string& input) {
        std::string s = input;
        std::ranges::transform(s, s.begin(), ::tolower);
        return s;
    }

    void TextInputForm::MoveCursorLeft() {
        if (cursor_position > 0)
            cursor_position--;
    }

    void TextInputForm::MoveCursorRight() {
        if (cursor_position < CursorMaxPosition()-1)
            cursor_position++;
    }

    unsigned int TextInputForm::CursorMaxPosition() const {
        return input_buffer.length();
    }

    void TextInputForm::SetCursorPosition(unsigned int pos) {
        pos = Math::Clamp<unsigned int>(pos, 0, CursorMaxPosition());
        cursor_position = pos;
    }

    void TextInputForm::SendInput(bool clear_input) {
        OnReturn.Invoke(input_buffer);

        if (clear_input) {
            input_buffer = "";
            cursor_position = 0;
        }
    }

    /// Returns true if the given key is a 'special' key, which should not be typed into the buffer, and has no other special function.
    bool IsNonAlphanumerosymbolic(const Key& key) {
        return key == Keys::LeftShift || key == Keys::RightShift || key == Keys::LeftControl || key == Keys::RightControl
        || key == Keys::LeftAlt || key == Keys::RightAlt || key == Keys::Super || key == Keys::Escape
        || key == Keys::F1 || key == Keys::F2 || key == Keys::F3 || key == Keys::F4 || key == Keys::F5 || key == Keys::F6
        || key == Keys::F7 || key == Keys::F8 || key == Keys::F9 || key == Keys::F10 || key == Keys::F11 || key == Keys::F12;
    }

    void TextInputForm::Copy() {}
    void TextInputForm::Paste() {}
    void TextInputForm::Cut() {}
    void TextInputForm::Delete() { }
    void TextInputForm::Backspace() {
        if (cursor_position > 0) {
            input_buffer = input_buffer.erase(cursor_position-1, 1);
            cursor_position--;
        }
    }
    void TextInputForm::PushStringToCurrentPlaceInInputBuffer(const std::string& snippet) {
        input_buffer = input_buffer.insert(cursor_position, snippet);
        cursor_position += snippet.length();
    }
    void TextInputForm::PushKeyToCurrentPlaceInInputBuffer(const Key& key) {
        std::string insertion = lowercase(key.Mnemonic);
        if (InputService::IsKeyDown(Keys::LeftShift) || InputService::IsKeyDown(Keys::RightShift)) {
            insertion = uppercase(insertion);
            // characters that don't work with std::toupper, so we do it manually.
            if (key == Keys::One) insertion = "!";
            if (key == Keys::Two) insertion = "@";
            if (key == Keys::Three) insertion = "#";
            if (key == Keys::Four) insertion = "$";
            if (key == Keys::Five) insertion = "%";
            if (key == Keys::Six) insertion = "&";
            if (key == Keys::Seven) insertion = "'";
            if (key == Keys::Eight) insertion = "*";
            if (key == Keys::Nine) insertion = "(";
            if (key == Keys::Zero) insertion = ")";
        }
        PushStringToCurrentPlaceInInputBuffer(insertion);
    }

    bool TextInputForm::ObserveKeyInput(Key key, bool pressed) {
        Widget::ObserveKeyInput(key, pressed);
        if (!pressed || !focused) return false;

        if (key == Keys::Return || key == Keys::NumPadReturn) {
            SendInput(clear_text_on_return);
            return true;
        }

        if (key == Keys::LeftArrow) {
            MoveCursorLeft();
            return true;
        }

        if (key == Keys::RightArrow) {
            MoveCursorRight();
            return true;
        }

        if (InputService::IsKeyDown(Keys::LeftControl)) {
            if (key == Keys::C) { Copy(); return true; }
            if (key == Keys::V) { Paste(); return true; }
            if (key == Keys::X) { Cut(); return true; }
        }

        if (IsNonAlphanumerosymbolic(key)) return false;
        if (blacklist.contains(key.Mnemonic)) return false;
        if (key == Keys::Backspace) {Backspace(); return true; }

        // No other condition met, so we assume we should push this key into the input buffer.
        PushKeyToCurrentPlaceInInputBuffer(key);
        return true;
    }

    bool TextInputForm::HasFocus() const { return focused; }

    void TextInputForm::SetFocused(bool focused) { this->focused = focused;}

    void TextInputForm::GrabFocus() { SetFocused(true); }

    void TextInputForm::DropFocus() { SetFocused(false); }

    bool TextInputForm::ClearTextOnReturn() const { return clear_text_on_return; }

    void TextInputForm::ClearTextOnReturn(bool value) { clear_text_on_return = value;}

    bool TextInputForm::DropFocusOnReturn() const { return drop_focus_on_return;}

    void TextInputForm::DropFocusOnReturn(bool value) { drop_focus_on_return = value;}

    std::set<std::string> TextInputForm::GetBlacklist() const { return blacklist;}

    std::string TextInputForm::InputBuffer() const { return input_buffer;}

    void TextInputForm::SetInputBuffer(const std::string &value) { input_buffer = value;}

    void TextInputForm::ClearInputBuffer() { SetInputBuffer(""); }

    void TextInputForm::SetBlacklist(const std::set<std::string> &value) {
        blacklist = value;
    }

    void TextInputForm::AddToBlacklist(const std::string& value) {
        blacklist.insert(value);
    }

}
